'''
RICInterface
'''
from typing import Callable, Dict, Union, Optional
import time
import threading
import logging
import json
import os
from .RICProtocols import DecodedMsg, RICProtocols
from .RICCommsBase import RICCommsBase
from .RateAverager import RateAverager
from .ValueAverager import ValueAverager
from .Exceptions import MartyTransferException

logger = logging.getLogger(__name__)

class RICInterface:
    '''
    RICInterface
    '''
    def __init__(self, commsHandler: RICCommsBase) -> None:
        '''
        Initialise RICInterface
        '''
        self.commsHandler = commsHandler
        self.ricProtocols = RICProtocols()
        self.decodedMsgCB = None
        self.logLineCB = None
        # Message command/response matching
        self.msgTimerCB = None
        self._msgsOutstanding: Dict = {}
        self._msgsOutstandingLock = threading.Lock()
        self.msgRespTimeoutSecs = 1.5
        # AddOn QueryRaw message matching
        self._rawQueryOutstanding: Dict = {}
        self._rawQueryOutstandingLock = threading.Lock()
        self._rawQueryMsgKey = 1
        # Stats
        self.msgRxRate = RateAverager()
        self.msgTxRate = RateAverager()
        self.roundTripInfo = ValueAverager()
        self.statsMatched = 0
        self.statsUnMatched = 0
        self.statsUnNumbered = 0
        self.statsTimedOut = 0
        self.uploadBytesPerSec = ValueAverager()
        # File send vars
        self._fileOTAStartFailed = False
        self._fileOTAStartedOK = False
        self._fileNotStarted = False
        self._fileUserCancel = False
        self._fileSendOkTo = 0
        self._fileSendNewOkTo = False
        self._fileFailedInFirmware = False
        self.BLOCK_ACK_TIMEOUT = 15
        self.BATCH_RETRY_MAX = 3

    def __del__(self) -> None:
        self.commsHandler.close()

    def open(self, openParams: Dict) -> bool:
        '''
        Open interface to RIC
        Args:
            openParams: dict containing params used to open the connection
                    (see RICCommsBase and derived classes for details)
        Returns:
            True if open succeeded or port is already open
        Throws:
            SerialException: if the serial port cannot be opened
        '''
        self.commsHandler.setRxFrameCB(self._onRxFrameCB)
        self.commsHandler.setRxLogLineCB(self._onLogLineCB)
        openOk = self.commsHandler.open(openParams)

        # Set default timeout based on interface
        self.msgRespTimeoutSecs = self.commsHandler.getMsgRespTimeoutSecs(self.msgRespTimeoutSecs)

        # Start timer to check message completion
        self.msgTimeoutCheckTimer = threading.Timer(1.0, self._msgTimeoutCheck)
        self.msgTimeoutCheckTimer.daemon = True
        self.msgTimeoutCheckTimer.start()
        return openOk

    def close(self) -> None:
        '''
        Close interface to RIC
        Args:
            none
        Returns:
            None
        '''
        self.commsHandler.close()

    def isOpen(self) -> None:
        return self.commsHandler.isOpen()

    def setDecodedMsgCB(self, onDecodedMsg: Callable[[DecodedMsg, 'RICInterface'], None]) -> None:
        '''
        Set callback on decoded message received from RIC
        Args:
            cb: callback function (takes 2 parameters: decoded message and this object)
        Returns:
            None
        '''
        self.decodedMsgCB = onDecodedMsg

    def setLogLineCB(self, onLogLine: Callable[[str], None]) -> None:
        '''
        Set callback on logging line received
        Args:
            onLogLine: callback function (takes 1 parameter which is the line of logging information)
        Returns:
            None
        '''
        self.logLineCB = onLogLine

    def setTimerCB(self, onMsgTimerCB: Callable[[], None]) -> None:
        '''
        Set callback which can be used to check for message types (e.g. publish)
        Args:
            onMsgTimerCB: callback function (takes 0 parameters)
        Returns:
            None
        '''
        self.msgTimerCB = onMsgTimerCB

    def sendRICRESTURL(self, msg: str, timeOutSecs: Optional[float] = None) -> bool:
        '''
        Send RICREST URL message
        Args:
            msg: string containing command URL
            timeOutSecs: message time-out override in seconds (or None to use default)
        Returns:
            True if message sent
        '''
        ricRestMsg, msgNum = self.ricProtocols.encodeRICRESTURL(msg)
        timeOutSecs = timeOutSecs if timeOutSecs is not None else self.msgRespTimeoutSecs
        with self._msgsOutstandingLock:
            self._msgsOutstanding[msgNum] = {"timeSent": time.time(), "timeOutSecs": timeOutSecs}
        # logger.debug(f"sendRICRESTURL msgNum {msgNum} time {time.time()} msg {msg}")
        self.commsHandler.send(ricRestMsg)
        self.msgTxRate.addSample()
        return True

    def cmdRICRESTURLSync(self, msg: str, timeOutSecs: Optional[float] = None) -> Dict:
        '''
        Send RICREST URL message and wait for response
        Args:
            msg: string containing command URL
            timeOutSecs: message time-out override in seconds (or None to use default)
        Returns:
            Response turned into a dictionary (from JSON)
        '''
        ricRestMsg, msgNum = self.ricProtocols.encodeRICRESTURL(msg)
        # logger.debug(f"msgNum {msgNum} msg {msg}")
        msgSendTime = time.time()
        timeOutSecs = timeOutSecs if timeOutSecs is not None else self.msgRespTimeoutSecs
        # logger.debug(f"cmdRICRESTURLSync msgNum {msgNum} timeout {timeOutSecs} msg {msg}")
        with self._msgsOutstandingLock:
            self._msgsOutstanding[msgNum] = {"timeSent": msgSendTime, "timeOutSecs": timeOutSecs, "awaited":True}
        self.commsHandler.send(ricRestMsg)
        self.msgTxRate.addSample()
        # Wait for result
        while time.time() < msgSendTime + timeOutSecs:
            with self._msgsOutstandingLock:
                # Should be an outstanding message - if not there's a problem
                if msgNum not in self._msgsOutstanding:
                    logger.debug(f"sendRICRESTURLSync msgNum {msgNum} not in _msgsOutstanding")
                    return {"rslt":"failResponse"}
                # Check if response received
                if self._msgsOutstanding[msgNum].get("respValid", False):
                    try:
                        # Get response
                        respStr = self._msgsOutstanding[msgNum].get("resp", None)
                        respObj = {}
                        if respStr:
                            respObj = json.loads(respStr.payload.rstrip('\0'))
                        debugMsgRespTime = self._msgsOutstanding[msgNum].get("respTime", 0)
                        # logger.debug(f"sendRICRESTURLSync msgNum {msgNum} msg {msg} resp {json.dumps(respObj)} sendTime {msgSendTime} respTime {debugMsgRespTime}")
                        return respObj
                    except Exception as excp:
                        logger.warn(f"sendRICRESTURLSync msgNum {msgNum} response is not JSON {excp}")
            time.sleep(0.01)
        # Debug - if we get here we timed out
        logger.debug(f"sendRICRESTURLSync msgNum {msgNum} failTimeout")
        return {"rslt":"failTimeout"}

    def cmdRICRESTRslt(self, msg: str, timeOutSecs: Optional[float] = None) -> bool:
        '''
        Send RICREST URL message and wait for response
        Args:
            msg: string containing command URL
            timeOutSecs: message time-out override in seconds (or None to use default)
        Returns:
            Response turned into a dictionary (from JSON)
        '''
        response = self.cmdRICRESTURLSync(msg, timeOutSecs)
        return response.get("rslt", "") == "ok"

    def sendRICRESTCmdFrame(self, msg: Union[str,bytes], 
                    payload: Union[bytes, str, None] = None, timeOutSecs: Optional[float] = None) -> bool:
        '''
        Send RICREST command frame message
        Args:
            msg: string or bytes containing command frame
            payload: optional payload
            timeOutSecs: message time-out override in seconds (or None to use default)
        Returns:
            True if message sent
        '''
        ricRestMsg, msgNum = self.ricProtocols.encodeRICRESTCmdFrame(msg, payload)
        # logger.debug(f"sendRICRESTCmdFrame msgNum {msgNum} len {len(ricRestMsg)} msg {msg}")
        timeOutSecs = timeOutSecs if timeOutSecs is not None else self.msgRespTimeoutSecs
        with self._msgsOutstandingLock:
            self._msgsOutstanding[msgNum] = {"timeSent": time.time(), "timeOutSecs": timeOutSecs}
        self.commsHandler.send(ricRestMsg)
        self.msgTxRate.addSample()
        return True

    def sendRICRESTCmdFrameSync(self, msg: Union[str,bytes], 
                    payload: Union[bytes, str] = None,
                    timeOutSecs: Optional[float] = None) -> Dict:
        '''
        Send RICREST command frame message and wait for response
        Args:
            msg: string or bytes containing command frame
            payload: optional payload
            timeOutSecs: message time-out override in seconds (or None to use default)
        Returns:
            Response turned into a dictionary (from JSON)
        '''
        # Encode frame
        ricRestMsg, msgNum = self.ricProtocols.encodeRICRESTCmdFrame(msg, payload)
        # logger.debug(f"sendRICRESTCmdFrameSync msgNum {msgNum} len {len(ricRestMsg)} msg {msg}")
        msgSendTime = time.time()
        timeOutSecs = timeOutSecs if timeOutSecs is not None else self.msgRespTimeoutSecs
        with self._msgsOutstandingLock:
            self._msgsOutstanding[msgNum] = {"timeSent": msgSendTime, "timeOutSecs": timeOutSecs, "awaited":True}
        self.commsHandler.send(ricRestMsg)
        self.msgTxRate.addSample()
        # Wait for result
        while time.time() < msgSendTime + timeOutSecs:
            logger.debug(f"sendRICRESTCmdFrameSync while loop")
            with self._msgsOutstandingLock:
                # Should be an outstanding message - if not there's a problem
                if msgNum not in self._msgsOutstanding:
                    return {"rslt":"failResponse"}
                # Check if response received
                if self._msgsOutstanding[msgNum].get("respValid", False):
                    try:
                        # Get response
                        respStr = self._msgsOutstanding[msgNum].get("resp", None)
                        respObj = {}
                        if respStr:
                            respObj = json.loads(respStr.payload.rstrip('\0'))
                        debugMsgRespTime = self._msgsOutstanding[msgNum].get("respTime", 0)
                        # logger.debug(f"sendRICRESTCmdFrameSync msgNum {msgNum} msg {msg} resp {json.dumps(respObj)} sendTime {msgSendTime} respTime {debugMsgRespTime}")
                        return respObj
                    except Exception as excp:
                        logger.debug(f"msgNum {msgNum} response is not JSON {excp}", )
            time.sleep(0.01)
        logger.debug(f"sendRICRESTCmdFrameSync msgNum {msgNum} msg {msg} sendTime {msgSendTime} timeNow {time.time()} timeout {timeOutSecs}")
        return {"rslt":"failTimeout"}

    def sendRICRESTFileBlock(self, data: bytes) -> bool:
        '''
        Send RICREST file block
        Args:
            data: bytes of file data
        Returns:
            True if message sent
        '''
        ricRestMsg, _ = self.ricProtocols.encodeRICRESTFileBlock(data)
        # logger.debug(f"sendRICRESTFileBlock msgNum {msgNum} len {len(ricRestMsg)}")
        self.commsHandler.send(ricRestMsg)
        self.msgTxRate.addSample()
        return True

    def newRoundTrip(self, rtTime:int) -> None:
        '''
        Indicate a new round-trip for a message is complete - this is
        for statistics gathering
        Args:
            rtTime: time taken for round-trip in seconds
        Returns:
            None
        '''
        self.roundTripInfo.add(rtTime)
        # logger.debug(f"RTTime {self.roundTripInfo.getAvg()}")

    def sendTestMsgs(self, numMsgs:int, bytesPerMsg: int) -> None:
        '''
        Send a batch of test messages
        Args:
            numMsgs: number of messages to send
            bytesPerMsg: bytes in each message
        Returns:
            None
        '''
        dataBlock = bytearray(bytesPerMsg)
        for i in range(numMsgs):
            self.commsHandler.send(dataBlock)
            self.msgTxRate.addSample()

    def _sendFileProgressCheckAbort(self, progressCB: Callable[[int, int, 'RICInterface'], bool], 
                    currentPos: int, fileSize: int) -> bool:
        if self._fileOTAStartFailed or self._fileNotStarted or self._fileUserCancel or self._fileFailedInFirmware:
            return True
        if not progressCB:
            return False
        if not progressCB(currentPos, fileSize, self):
            self.sendRICRESTCmdFrameSync('{"cmdName":"ufCancel"}')
            return True
        return False

    def sendFile(self, filename: str, fileDest: str, reqStr: str = '', 
                progressCB: Callable[[int, int, 'RICInterface'], bool] = None) -> bool:
        '''
        Send a file (from the file system)
        Args:
            filename: name of file to send
            fileDest: "fs" to upload to file system, "ricfw" for new RIC firmware
            reqStr: API request used for transfer, if left blank this is inferred from fileDest, other
                    values include "fileupload" and "espfwupdate" - see RIC documentation for API
            progressCB: callback used to indicate how file send is progressing, callable takes three params
                    which are bytesSent, totalBytes and the interface to RIC (of type RICInterface) and
                    returns a bool which should be True to continue the file upload or False to abort
        Returns:
            True if operation succeeded
        Throws:
            OSError: operating system exceptions
            May throw other exceptions so include a general exception handler
        '''
        # Send new firmware in the bin folder using the CommandHandler protocol
        with open(filename, "rb") as f:

            # Read firmware
            binaryImage = f.read()
            binaryImageLen = len(binaryImage)
            # logger.debug(f"File {filename} is {binaryImageLen} bytes long")

            # Check if we're uploading firmware
            isFirmware = fileDest == "ricfw"
            if reqStr == '':
                reqStr = 'espfwupdate' if isFirmware else 'fileupload'
            uploadName = "fw" if isFirmware else os.path.basename(filename)
            self._fileOTAStartFailed = False
            self._fileOTAStartedOK = False
            self._fileNotStarted = False
            self._fileUserCancel = False
            self._fileFailedInFirmware = False

            # Frames follow the approach used in the web interface start, block..., end
            resp = self.sendRICRESTCmdFrameSync('{' + f'"cmdName":"ufStart","reqStr":"{reqStr}","fileType":"{fileDest}",' + \
                            f'"fileName":"{uploadName}","fileLen":{str(binaryImageLen)}' + '}', 
                            timeOutSecs = 10)
            if resp.get("rslt","") != "ok":
                raise MartyTransferException("File transfer start not acknowledged")

            # Block and batch sizes
            blockMaxSize = self.commsHandler.commsParams.fileTransfer.get("fileBlockMax", 5000)
            blockMaxSize = resp.get("batchMsgSize", blockMaxSize)
            batchAckSize = resp.get("batchAckSize", 50)
            self._fileSendOkTo = 0

            # Debug
            # logger.debug(f"ricIF sendFile blockMaxSize {blockMaxSize} batchAckSize {batchAckSize} resp {resp}")

            # Progress and check for abort
            if self._sendFileProgressCheckAbort(progressCB, self._fileSendOkTo, binaryImageLen):
                return False

            # Wait for a period depending on whether we're sending firmware - this is because starting
            # a firmware update involves the ESP32 in a long-running activity and the firmware becomes
            # unresponsive during this time
            if isFirmware:
                for i in range(5):
                    time.sleep(1)
                    if self._sendFileProgressCheckAbort(progressCB, 0, binaryImageLen):
                        return False
                    if self._fileOTAStartedOK:
                        break

            # Debug
            # logger.debug(f"ricIF sendFile starting to send file data ...")

            # Send file blocks
            numBlocks = 0
            batchRetryCount = 0
            while self._fileSendOkTo < binaryImageLen:

                # NOTE: first batch MUST be of size 1 (not batchAckSize) because RIC performs a long-running
                # blocking task immediately after receiving the first message in a firmware
                # update - although this could be relaxed for non-firmware update file uploads
                sendFromPos = self._fileSendOkTo
                batchStartPos = self._fileSendOkTo
                batchStartTime = time.time()
                batchSize = 1 if sendFromPos == 0 else batchAckSize
                batchBlockIdx = 0
                self._fileSendNewOkTo = False
                while batchBlockIdx < batchSize and sendFromPos < binaryImageLen:

                    # Send block
                    blockToSend = binaryImage[sendFromPos:sendFromPos+blockMaxSize]
                    self.sendRICRESTFileBlock(sendFromPos.to_bytes(4, 'big') + blockToSend)
                    sendFromPos += blockMaxSize
                    batchBlockIdx += 1

                    # Check if we have received an error
                    if self._fileNotStarted or self._fileUserCancel or self._fileFailedInFirmware:
                        break

                # Debug
                # logger.debug(f"ricIF sendFile sent batch - start at {batchStartPos} end at {sendFromPos} okto {self._fileSendOkTo}")

                # Wait for response (there is a timeout at the ESP end to ensure a response is always returned
                # even if blocks are dropped on reception at ESP) - the timeout here is for these responses
                # being dropped
                timeNow = time.time()
                while time.time() - timeNow < self.BLOCK_ACK_TIMEOUT:
                    # Progress update
                    if self._sendFileProgressCheckAbort(progressCB, self._fileSendOkTo, binaryImageLen):
                        return False

                    # Debug
                    # logger.debug(f"ricIF sendFile checking for OKTO {self._fileSendOkTo} batchStartPos {batchStartPos}")

                    # Check for okto
                    if self._fileSendNewOkTo:
                        batchRetryCount = 0
                        # Update stats
                        if self._fileSendOkTo > batchStartPos:
                            batchBytesPerSec = (self._fileSendOkTo - batchStartPos) / (time.time() - batchStartTime)
                            self.uploadBytesPerSec.add(batchBytesPerSec)
                        break

                    # Wait
                    time.sleep(1)

                # Check if no okto has been received with a greater position than batchStartPos
                if self._fileSendOkTo <= batchStartPos:
                    batchRetryCount += 1
                    if batchRetryCount > self.BATCH_RETRY_MAX:
                        return False

                # Block count
                numBlocks += 1

            # Debug
            # logger.debug(f"ricIF sendFile sending END")

            # End frame
            resp = self.sendRICRESTCmdFrameSync('{' + f'"cmdName":"ufEnd","reqStr":"fileupload","fileType":"{fileDest}",' + \
                            f'"fileName":"{filename}","fileLen":{str(binaryImageLen)},' + \
                            f'"blockCount":{str(numBlocks)}' + '}', 
                            timeOutSecs = 5)
            if resp.get("rslt","") != "ok":
                return False
            return True

    def getStats(self) -> Dict:
        return {
            "roundTripAvgMS":self.roundTripInfo.getAvg()*1000,
            "msgRxRatePS":self.msgRxRate.getAvg(),
            "msgTxRatePS":self.msgTxRate.getAvg(),
            "unmatched":self.statsUnMatched,
            "matched":self.statsMatched,
            "unnumbered":self.statsUnNumbered,
            "timedOut":self.statsTimedOut,
            "uploadBPS":self.uploadBytesPerSec.getAvg(),
            "rxCount":self.msgRxRate.getTotal(),
            "txCount":self.msgTxRate.getTotal(),
        }

    def addOnQueryRaw(self, addOnName: str, dataToWrite: bytes, numBytesToRead: int,
                        timeOutSecs: Optional[float] = None) -> Dict:
        '''
        Write and read an addOn directly (raw-mode)
        Args:
            addOnName, name of the addOn (see get_add_ons_status() at the top level or response to
                addon/list REST API command)
            dataToWrite, can be zero length if nothing is to be written, the first byte will generally
                be the register or opcode of the addOn
            numBytesToRead: number of bytes to read from the device - can be zero
            timeOutSecs: message time-out override in seconds (or None to use default)
        Returns:
            Dict with keys including:
                "rslt" - the result which will be "ok" if the query succeeded
                "dataRead" - the data read from the addOn
        '''
        # Handle msgKey
        msgKey = self._rawQueryMsgKey
        self._rawQueryMsgKey += 1
        if self._rawQueryMsgKey > 99999:
            self._rawQueryMsgKey = 1

        # Form command
        hexWrStr = dataToWrite.hex()
        ricRestCmd = f"elem/{addOnName}/json?cmd=raw&hexWr={hexWrStr}&numToRd={numBytesToRead}&msgKey={msgKey}"

        # Register in raw message matching
        timeOutSecs = timeOutSecs if timeOutSecs is not None else self.msgRespTimeoutSecs
        with self._rawQueryOutstandingLock:
            self._rawQueryOutstanding[msgKey] = {"timeSent": time.time(), "timeOutSecs": timeOutSecs, "awaited":True}

        # Send message
        resp = self.cmdRICRESTURLSync(ricRestCmd)
        resp["dataRead"] = b""
        if resp.get("rslt", "") != "ok":
            self._rawQueryOutstanding.pop(msgKey)
            return resp

        # Wait for a report message generated by the addOn access process
        msgSendTime = time.time()
        while time.time() < msgSendTime + timeOutSecs:
            with self._rawQueryOutstandingLock:
                # Should be an outstanding message - if not there's a problem
                if msgKey not in self._rawQueryOutstanding:
                    resp["rslt"] = "failReport"
                    return resp
                # Check if a report received
                if self._rawQueryOutstanding[msgKey].get("reptValid", False):
                    # Get report
                    reptObj = self._rawQueryOutstanding[msgKey].get("reptObj", {})
                    # logger.debug(f"msgKey {msgKey} msg {ricRestCmd} rept {json.dumps(reptObj)} sendTime {msgSendTime}")
                    resp["dataRead"] = reptObj.get("hexRd", b"")
                    return resp
            time.sleep(0.01)
        return {"rslt":"failTimeout"}

    def _onRxFrameCB(self, frame: bytes) -> None:
        # logger.debug(f"_onRxFrameCB Rx len {len(frame)} data {frame.hex()}")
        self.msgRxRate.addSample()
        decodedMsg = self.ricProtocols.decodeRICFrame(frame)
        doRxCallback = True
        if decodedMsg.msgNum != 0:
            # Numbered message - this is the response to a REST API command
            # logger.debug(f"_onRxFrameCB msgNum {decodedMsg.msgNum} {decodedMsg.payload}")
            isUnmatched = False
            with self._msgsOutstandingLock:
                if decodedMsg.msgNum in self._msgsOutstanding:
                    msgRec = self._msgsOutstanding[decodedMsg.msgNum]
                    roundTripTime = time.time() - msgRec["timeSent"]
                    self.newRoundTrip(roundTripTime)
                    if not msgRec.get("awaited", False):
                        self._msgsOutstanding.pop(decodedMsg.msgNum)
                    else:
                        msgRec["resp"] = decodedMsg
                        msgRec["respValid"] = True
                        msgRec["respTime"] = time.time()
                    self.statsMatched += 1
                else:
                    isUnmatched = True
                    self.statsUnMatched += 1
            if isUnmatched:
                logger.warn(f"_onRxFrameCB Unmatched msgNum {decodedMsg.msgNum}")
            doRxCallback = isUnmatched
        else:
            # logger.debug(f"_onRxFrameCB unnumbered restType {decodedMsg.restType} msgType {decodedMsg.msgTypeCode}")
            if decodedMsg.msgTypeCode == RICProtocols.MSG_TYPE_REPORT:
                # Report message - this can include results of accessing addOns
                # logger.debug(f"_onRxFrameCB REPORT {decodedMsg.payload}")
                reptObj = {}
                try:
                    reptObj = json.loads(decodedMsg.payload.rstrip('\0'))
                except Exception as excp:
                    logger.warn(f"_onRxFrameCB REPORT is not JSON {excp}")
                msgKey = reptObj.get("msgKey", '')
                if type(msgKey) is str:
                    try:
                        msgKey = int(msgKey)
                    except:
                        msgKey = 0
                if msgKey != 0:
                    isUnmatched = False
                    with self._rawQueryOutstandingLock:
                        if msgKey in self._rawQueryOutstanding:
                            msgRec = self._rawQueryOutstanding[msgKey]
                            if not msgRec.get("awaited", False):
                                self._rawQueryOutstanding.pop(msgKey)
                            else:
                                msgRec["reptObj"] = reptObj
                                msgRec["reptValid"] = True
                        else:
                            isUnmatched = True
                    if isUnmatched:
                        logger.debug(f"Unmatched msgKey {msgKey}")
                    doRxCallback = isUnmatched
            elif decodedMsg.msgTypeCode == RICProtocols.MSG_TYPE_RESPONSE:
                # logger.debug(f"RESPONSE {decodedMsg.payload}")
                # Check for okto message
                reptObj = {}
                try:
                    reptObj = json.loads(decodedMsg.payload.rstrip('\0'))
                except Exception as excp:
                    logger.warn(f"_onRxFrameCB REPORT is not JSON {excp}")
                if "okto" in reptObj:
                    okto = reptObj.get("okto", -1)
                    if self._fileSendOkTo < okto:
                        self._fileSendOkTo = okto
                    self._fileSendNewOkTo = True
                    # logger.debug(f"OKTO MESSAGE {reptObj['okto']}")
                elif "cmdName" in reptObj:
                    cmdName = reptObj.get("cmdName","")
                    if cmdName == "ufBlock" or cmdName == "ufStatus" or cmdName == "ufCancel":
                        reason = reptObj.get("reason","")
                        if reason == "OTAStartFailed":
                            self._fileOTAStartFailed = True
                        elif reason == "OTAStartedOK":
                            self._fileOTAStartedOK = True
                        elif reason == "notStarted":
                            self._fileNotStarted = True
                        elif reason == "userCancel":
                            self._fileUserCancel = True
                        elif reason == "failRetries" or reason == "failTimeout":
                            self._fileFailedInFirmware = True
                    logger.warn(f"_onRxFrameCB {cmdName} reason {reason}")
                else:
                    logger.warn(f"_onRxFrameCB response not OkTo or fileUpload ... {decodedMsg.payload}")

            self.statsUnNumbered += 1
        # Call the decoded-message callback with the message
        if doRxCallback and self.decodedMsgCB is not None:
            self.decodedMsgCB(decodedMsg, self)

    def _onLogLineCB(self, line: str) -> None:
        if self.logLineCB:
            self.logLineCB(line)

    def _msgTimeoutCheck(self) -> None:
        # Check messages outstanding
        # logger.debug("Check outstanding messages")
        msgIdxsToRemove = []
        msgIdxsTimedOut = []
        with self._msgsOutstandingLock:
            for msgItem in self._msgsOutstanding.items():
                timeSinceSent = time.time() - msgItem[1].get("timeSent", 0)
                if timeSinceSent > msgItem[1].get("timeOutSecs", self.msgRespTimeoutSecs):
                    if not msgItem[1].get("respValid", False):
                        msgIdxsTimedOut.append(msgItem[0])
                    msgIdxsToRemove.append(msgItem[0])

        # Hint to comms layer if messages are failing
        if len(msgIdxsTimedOut) > 0:
            self.commsHandler.hintMsgTimeout(len(msgIdxsTimedOut))

        # Debug
        for msgIdx in msgIdxsTimedOut:
            logger.warn(f"Message {msgIdx} timed out timeSent {self._msgsOutstanding[msgIdx].get('timeSent',0)} timeNow {time.time()}")

        # Remove timed-out messages
        for msgIdx in msgIdxsToRemove:
            with self._msgsOutstandingLock:
                self._msgsOutstanding.pop(msgIdx)

        # Check rawQuery reports outstanding
        # logger.debug("Check rawQuery reports outstanding")
        msgKeysToRemove = []
        msgKeysTimedOut = []
        with self._rawQueryOutstandingLock:
            for msgItem in self._rawQueryOutstanding.items():
                timeSinceSent = time.time() - msgItem[1].get("timeSent", 0)
                if timeSinceSent > msgItem[1].get("timeOutSecs", self.msgRespTimeoutSecs):
                    if not msgItem[1].get("reptValid", False):
                        msgKeysTimedOut.append(msgItem[0])
                    msgKeysToRemove.append(msgItem[0])

        # Debug
        for msgKey in msgKeysTimedOut:
            logger.warn(f"rawQuery {msgKey} timed out at time {time.time()}")
            self.statsTimedOut += 1

        # Remove timed-out rawQueries
        for msgKey in msgKeysToRemove:
            with self._rawQueryOutstandingLock:
                self._rawQueryOutstanding.pop(msgKey)

        # Restart timer
        if self.commsHandler is not None and self.commsHandler.isOpen():
            self.msgTimeoutCheckTimer = threading.Timer(1.0, self._msgTimeoutCheck)
            self.msgTimeoutCheckTimer.start()

        # Callback on timer if required
        if self.msgTimerCB:
            self.msgTimerCB()

    def getTestOutput(self) -> dict:
        return self.commsHandler.getTestOutput()



from typing import Callable, Dict, List, Optional, Tuple, Union, Tuple
import sys
import time
import socket
import struct
from warnings import warn
from .ClientGeneric import ClientGeneric
from .Exceptions import (MartyConnectException,
                         MartyCommandException,
                         ArgumentOutOfRangeException,
                         UnavailableCommandException)

class ClientMV1(ClientGeneric):
    '''
    Lower level interface class between the `Marty` abstracted
    control class and the RIC socket interface
    '''
    SOCKET_PORT = 24

    SIDE_CODES = {
        'left'    : '\x00',
        'right'   : '\x01',
        'forward' : '\x02',
        'back'    : '\x03',
        'auto'    : '\xff',
    }

    EYE_POSES = {
        'angry': -15,
        'excited': 30,
        'normal': 0,
        'wide': 90
    }

    def __init__(self, method: str, locator: str,
                port:str = None, timeout:float = 5.0,
                debug:bool = False,
                default_lifelike: bool = True,
                *args, **kwargs):
        '''
        Initialise connection to remote Marty over a IPv4 socket by name 'loc' over port 24
        Args:
            loc, type str, must either resolve to an IP or be an IP address
        Raises:
            MartyConnectException if the socket failed to make the connection to the host
        '''
        if 'blocking' not in kwargs or kwargs['blocking'] is None:
            warn('Since martypy 3.0.0, all movement commands are blocking by default. '
                 'To go back to the old default, add `blocking=False` to your Marty '
                 'constructor. To silence this warning, provide the `blocking` argument '
                 f'explicitly (e.g. `Marty("{method}", "{locator}", blocking=True)`)',
                 stacklevel=3)
        super().__init__(*args, **kwargs)

        if port is None:
            self.port = self.SOCKET_PORT
        else:
            self.port = port

        self.debug = debug

        self.loc = locator
        self.timeout = timeout
        self.default_lifelike = default_lifelike

        self.sock = self._socket_factory()

        self.COMMANDS_LUT = dict()
        self._init_commands()


    def start(self):
        # To be able to do anything:
        self.enable_safeties(True)
        self.enable_motors(True)

        if self.default_lifelike:
            self.lifelike_behaviour(True)

    def close(self):
        if self.sock:
            self.sock.close()
        self.sock = None

    def wait_if_required(self, expected_wait_ms: int, blocking_override: Union[bool, None]):
        if not self.is_blocking(blocking_override):
            return
        time.sleep(expected_wait_ms/1000)

    def hello(self):
        return self._execute('hello')

    def get_ready(self):
        return self._execute('hello')

    def discover(self):
        return self._discover()

    def stop(self, stop_type: str, stopCode: int) -> bool:
        M1_STOP_TYPE = ['\x00','\x01','\x02','\x03','\x04','\x05']
        return self._execute('stop', M1_STOP_TYPE[stopCode])

    def resume(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def hold_position(self, hold_time: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def move_joint(self, joint_id: int, position: int, move_time: int) -> bool:
        '''
        Move a specific joint to a position
        Args:
            move_time: how long this movement should last, in milliseconds
        '''
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('move_joint',
                                   self._pack_uint8(joint_id),
                                   self._pack_int8(position),
                                   dur_lsb, dur_msb)

    def get_joint_position(self, joint_id: Union[int, str]) -> float:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_joint_current(self, joint_name_or_num: Union[int, str]) -> float:
        if type(joint_name_or_num) is str:
            raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)
        return self.get_motor_current(int(joint_name_or_num))

    def get_joint_status(self, joint_name_or_num: Union[int, str]) -> int:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def lean(self, direction: str, amount: Optional[int], move_time: int) -> bool:
        if amount is None:
            amount = 50
        try:
            directionNum = self.SIDE_CODES[direction]
        except KeyError:
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(self.SIDE_CODES.keys()), direction))
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('lean', directionNum,
                                   self._pack_int8(amount),
                                   dur_lsb, dur_msb)

    def walk(self, num_steps: int = 2, start_foot:str = 'auto', turn: int = 0,
                step_length:int = 25, move_time: int = 1500) -> bool:
        try:
            side_c = self.SIDE_CODES[start_foot]
        except KeyError:
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(self.SIDE_CODES.keys()), start_foot))
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('walk',
                                   self._pack_uint8(num_steps),
                                   self._pack_int8(turn),
                                   dur_lsb, dur_msb,
                                   self._pack_int8(step_length),
                                   side_c)

    def eyes(self, joint_id: int, pose_or_angle: Union[str, int], move_time: int = 1000) -> bool:
        if type(pose_or_angle) is str:
            try:
                angle = self.EYE_POSES[pose_or_angle]
            except KeyError:
                raise MartyCommandException("pose must be one of {}, not '{}'"
                                            "".format(set(self.EYE_POSES.keys()), pose_or_angle))
        else:
            angle = pose_or_angle
        return self.move_joint(joint_id, angle, move_time)

    def kick(self, side: str = 'right', twist: int = 0, move_time: int = 2500) -> bool:
        side_c = self.SIDE_CODES[side]
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('kick', side_c,
                                   self._pack_int8(twist),
                                   dur_lsb, dur_msb)

    def arms(self, left_angle: int, right_angle: int, move_time: int) -> bool:
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('arms',
                                   self._pack_int8(right_angle),
                                   self._pack_int8(left_angle),
                                   dur_lsb, dur_msb)

    def celebrate(self, move_time: int = 4000) -> bool:
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('celebrate', dur_lsb, dur_msb)

    def circle_dance(self, side: str = 'right', move_time: int = 2500) -> bool:
        side_c = self.SIDE_CODES[side]
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('circle_dance',
                                   side_c,
                                   dur_lsb, dur_msb)

    def dance(self, side: str = 'right', move_time: int = 3000) -> bool:
        for i in range(2):
            self.circle_dance(side, int(move_time/3))
            self.arms(90, 90, int(move_time/6))
            self.arms(45, 45, int(move_time/6))
        return self.stand_straight(int(move_time/3))

    def wiggle(self, move_time: int = 4000) -> bool:
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('celebrate', dur_lsb, dur_msb)

    def stand_straight(self, move_time: int = 2000) -> bool:
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('stand_straight', dur_lsb, dur_msb)

    def sidestep(self, side: str, steps: int = 1, step_length: int = 35,
            move_time: int = 1000) -> bool:
        side_c = self.SIDE_CODES[side]
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('sidestep', side_c,
                                   self._pack_int8(steps),
                                   dur_lsb, dur_msb,
                                   self._pack_int8(step_length))

    def play_sound(self, freq_start: Union[float, str],
            freq_end: float,
            duration: int) -> bool:
        if type(freq_start) is str:
            raise MartyCommandException("first parameter must be a number for Marty V1")
        f_start_lsb, f_start_msb = self._pack_uint16(int(freq_start))
        f_end_lsb, f_end_msb = self._pack_uint16(int(freq_end))
        dur_lsb, dur_msb = self._pack_uint16(5000 if duration is None else duration)
        return self._execute('play_sound',
                                   f_start_lsb, f_start_msb,
                                   f_end_lsb, f_end_msb,
                                   dur_lsb, dur_msb)

    GPIO_PIN_MODES = {
        'digital in'  : '\x00',
        'analog in'   : '\x01',
        'digital out' : '\x02',
        #'servo'       : '\x03',
        #'pwm'         : '\x04',
    }

    def pinmode_gpio(self, gpio: int, mode: str) -> bool:
        return self._execute('gpio_mode', self._pack_uint8(gpio), self.GPIO_PIN_MODES[mode])

    def write_gpio(self, gpio: int, value: int) -> bool:
        byte_array = (self._pack_uint8(gpio),) + self._pack_float(value)
        return self._execute('gpio_write', *byte_array)

    def digitalread_gpio(self, gpio: int) -> bool:
        return bool(self._execute('gpio', self._pack_uint8(gpio)))

    def i2c_write(self, *byte_array: int) -> bool:
        return self._execute('i2c_write', *byte_array)

    def i2c_write_to_ric(self, address: int, byte_array: bytes) -> bool:
        data = ['\x1B']         #i2c_write opcode
        data.append(address)    #i2c address
        data.append(byte_array) #message
        data += self.ros_serial_formatter(111, *byte_array) #ros serial format
        return self._execute('i2c_write', *data)

    def get_battery_voltage(self) -> float:
        return self._execute('battery')

    def get_battery_remaining(self) -> float:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_distance_sensor(self) -> float:
        return self._execute('distance')

    def get_accelerometer(self, axis: str, axisCode: int) -> float:
        M1_ACCEL_AXES = ['\x00','\x01','\x02']
        if axis is None:
            raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)
        return self._execute('accel', M1_ACCEL_AXES[axisCode])

    def get_motor_current(self, motor_id: int) -> float:
        return self._execute('motorcurrent', str(motor_id))

    def enable_motors(self, enable: bool = True, clear_queue: bool = True) -> bool:
        if clear_queue:
            self.stop('clear queue', 0)
        if enable:
            return self._execute('enable_motors') and True
        else:
            return self._execute('disable_motors') and False

    def enable_safeties(self, enable: bool = True) -> bool:
        return self._execute('enable_safeties')

    def fall_protection(self, enable: bool = True) -> bool:
        return self._execute('fall_protection', enable)

    def motor_protection(self, enable: bool = True) -> bool:
        return self._execute('motor_protection', enable)

    def battery_protection(self, enable: bool = True) -> bool:
        return self._execute('battery_protection', enable)

    def buzz_prevention(self, enable: bool = True) -> bool:
        return self._execute('buzz_prevention', enable)

    def lifelike_behaviour(self, enable: bool = True) -> bool:
        return self._execute('lifelike_behaviour', enable)

    def set_parameter(self, *byte_array: int) -> bool:
        return self._execute('set_param', '\x1F', *byte_array)

    def save_calibration(self) -> bool:
        return self._execute('save_calibration')

    def clear_calibration(self) -> bool:
        return self._execute('clear_calibration')

    def is_calibrated(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def ros_command(self,  *byte_array: int) -> bool:
        return self._execute('ros_command', *byte_array)

    def keyframe (self, time: float, num_of_msgs: int, msgs: List) -> List[bytes]:
        processed_keyframe = []

        #Number of key frames
        len_byte0, len_byte1, len_byte2, len_byte3 = self._pack_int32(1)
        processed_keyframe.append(len_byte0)
        processed_keyframe.append(len_byte1)
        processed_keyframe.append(len_byte2)
        processed_keyframe.append(len_byte3)

        #Time (in seconds) to excute keyframe. This is float encoded.
        time_byte0, time_byte1, time_byte2, time_byte3 = self._pack_float(time)
        processed_keyframe.append(time_byte0)
        processed_keyframe.append(time_byte1)
        processed_keyframe.append(time_byte2)
        processed_keyframe.append(time_byte3)

        if(len(msgs) != num_of_msgs):
            raise MartyCommandException('Number of messages do not match entered messages')
        #Array length
        arr_len_byte0, arr_len_byte1, arr_len_byte2, arr_len_byte3 = self._pack_int32(num_of_msgs)
        processed_keyframe.append(arr_len_byte0)
        processed_keyframe.append(arr_len_byte1)
        processed_keyframe.append(arr_len_byte2)
        processed_keyframe.append(arr_len_byte3)

        #Messages
        for items in msgs:
            for values in items:
                processed_keyframe.append(self._pack_int8(values))

        return(processed_keyframe)

    def get_chatter(self) -> bytes:
        return self._execute('chatter')

    def get_firmware_version(self) -> bool:
        return self._execute('firmware_version')

    def _mute_serial(self) -> bool:
        return self._execute('mute_serial')

    def ros_serial_formatter(self, topicID: int, send: bool = False, *message: int) -> List[int]:
        msg = message

        msg_length = len(msg)
        #Message length in little endian format
        msg_length_LB = msg_length & 0xFF                           #3rd byte
        msg_length_HB = (msg_length >> 8) & 0xFF                    #4th byte

        checksum1 = 255 - ((msg_length_LB + msg_length_HB) % 256)   #5th byte

        #Topic ID in little endian format
        topic_ID_LB = topicID & 0xFF                                #6th byte
        topic_ID_HB = (topicID >> 8) & 0xFF                         #7th byte

        data_values_sum = 0
        for i in msg:
            data_values_sum += ord(i)

        checksum2 = 255 - ((topic_ID_LB + topic_ID_HB + data_values_sum) % 256) #final byte

        #encode into bytes
        command_to_be_sent = []
        command_to_be_sent += ('\xff',) #Sync Flag. Check ROS Wiki
        command_to_be_sent += ('\xfe',) #Protocol version. Check ROS Wiki
        command_to_be_sent += (chr(msg_length_LB),)
        command_to_be_sent += (chr(msg_length_HB),)
        command_to_be_sent += (chr(checksum1),)
        command_to_be_sent += (chr(topic_ID_LB),)
        command_to_be_sent += (chr(topic_ID_HB),)
        command_to_be_sent += msg
        command_to_be_sent += (chr(checksum2),)

        if(send == True):
            self.ros_command(*command_to_be_sent)

        return(command_to_be_sent)

    def _init_commands(self) -> None:

        # Extend basis LUT with specific handlers
        self._register_commands({
            'discover'           : self._discover,
            'battery'            : self._simple_sensor,
            'distance'           : self._simple_sensor,
            'accel'              : self._select_sensor,
            'motorcurrent'       : self._current_sensor,
            'gpio'               : self._select_sensor,
            'hello'              : self._fixed_command,
            'lean'               : self._fixed_command,
            'walk'               : self._fixed_command,
            'kick'               : self._fixed_command,
            'celebrate'          : self._fixed_command,
            'arms'               : self._fixed_command,
            'sidestep'           : self._fixed_command,
            'stand_straight'     : self._fixed_command,
            'circle_dance'       : self._fixed_command,
            'play_sound'         : self._fixed_command,
            'stop'               : self._fixed_command,
            'move_joint'         : self._fixed_command,
            'enable_motors'      : self._fixed_command,
            'disable_motors'     : self._fixed_command,
            'enable_safeties'    : self._fixed_command,
            'fall_protection'    : self._toggle_command,
            'motor_protection'   : self._toggle_command,
            'battery_protection' : self._toggle_command,
            'buzz_prevention'    : self._toggle_command,
            'lifelike_behaviour' : self._toggle_command,
            'clear_calibration'  : self._fixed_command,
            'save_calibration'   : self._fixed_command,
            'ros_command'        : self._command,
            'chatter'            : self._chatter,
            'set_param'          : self._command,
            'firmware_version'   : self._fixed_command,
            'mute_serial'        : self._fixed_command,
            'i2c_write'          : self._command,
            'gpio_write'         : self._fixed_command,
            'gpio_mode'          : self._fixed_command,
        })

        # for opcode, _ in self.COMMANDS_LUT.items():
        #     if opcode not in self.CMD_OPCODES.keys():
        #         print('Missing {}'.format(opcode))

    def _socket_factory(self) -> socket.socket:
        '''
        Generate a new socket connection
        '''
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            loc_ip = socket.gethostbyname(self.loc)
            sock.connect((loc_ip, self.port))
        except Exception as e:
            raise MartyConnectException(e)

        sock.settimeout(self.timeout)

        if sys.platform == 'linux':
            # Configure socket in keepalive mode
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # Time to wait before sending a keepalive
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 1)
            # Interval between keepalives
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 1)
            # Number of keepalive fails before declaring the connection broken
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 5)

        elif sys.platform == 'windows':
            # Keep alive, 5 sec timeout, 1 sec interval
            sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, 5000, 1000))

        return sock

    def __repr__(self):
        return '{}.{}{}'.format(self.__module__,
                             self.__class__.__name__,
                             self.sock.getsockname())

    @staticmethod
    def _discover(timeout: float = 10, *args: int, **kwargs: int) -> List[str]:
        '''
        Search for Marties on the network using a UDB multicast to port 4000
        '''
        socket_addr = "224.0.0.1"
        socket_port = 4000
        magic_command = b"AA"

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 32)
        sock.settimeout(timeout)
        sock.sendto(magic_command, (socket_addr, socket_port))

        found = []
        start = time.time()
        try:
            while (time.time() - start) < timeout:
                data, addr = sock.recvfrom(1000)
                found.append({addr: data})
        except socket.timeout:
            return found

    def is_moving(self) -> bool:
        '''
        Check if Marty is moving

        Args:
            none
        Returns:
            True if Marty is moving
        '''
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def is_paused(self) -> bool:
        '''
        Check if Marty is paused

        Args:
            none
        Returns:
            True if Marty is paused
        '''
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_robot_status(self) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_joints(self) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_power_status(self) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_add_ons_status(self) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_add_on_status(self, add_on_name_or_id: Union[int, str]) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def add_on_query(self, add_on_name: str, data_to_write: bytes, num_bytes_to_read: int) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_system_info(self) -> Dict:
        return {"HardwareVersion":"1.0", "SystemName":"MartyV1","SystemVersion":"1.0.0","SerialNo":"000001","MAC":"000000000000"}

    def set_marty_name(self, name: str) -> bool:
        return False

    def get_marty_name(self) -> str:
        return "Marty"

    def is_marty_name_set(self) -> bool:
        return False

    def get_hw_elems_list(self) -> List:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def send_ric_rest_cmd(self, ricRestCmd: str) -> None:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def send_ric_rest_cmd_sync(self, ricRestCmd: str) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def is_conn_ready(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def disco_off(self, add_on: str) -> bool :
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def disco_pattern(self, pattern: int, add_on: str) -> bool :
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)
        
    def disco_color(self, color: Union[str, Tuple[int, int, int]], add_on: str, region: Union[int, str]) -> bool:  
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def disco_group_operation(self, disco_operation: Callable, whoami_type_codes: set, operation_kwargs: dict) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    # Encodes Command Type flag, LSB size, MSB size, Data
    CMD_OPCODES = {
        'battery'            : ['\x01', '\x01', '\x00'],         # OK
        'distance'           : ['\x01', '\x08', '\x00'],         #
        'accel'              : ['\x01', '\x02'],                 # OK
        'motorcurrent'       : ['\x01', '\x03'],                 # OK
        'gpio'               : ['\x01', '\x04'],                 #
        'chatter'            : ['\x01', '\x05', '\x00',],        # Variable Length
        'hello'              : ['\x02', '\x01', '\x00', '\x00'], # OK
        'lean'               : ['\x02', '\x05', '\x00', '\x02'], #
        'walk'               : ['\x02', '\x07', '\x00', '\x03'], # OK
        'kick'               : ['\x02', '\x05', '\x00', '\x05'], # OK
        'celebrate'          : ['\x02', '\x03', '\x00', '\x08'], # OK
        'arms'               : ['\x02', '\x05', '\x00', '\x0B'], #
        'sidestep'           : ['\x02', '\x06', '\x00', '\x0E'], #
        'stand_straight'     : ['\x02', '\x03', '\x00', '\x0F'],
        'play_sound'         : ['\x02', '\x07', '\x00', '\x10'], #
        'stop'               : ['\x02', '\x02', '\x00', '\x11'], # OK
        'move_joint'         : ['\x02', '\x05', '\x00', '\x12'], #
        'enable_motors'      : ['\x02', '\x01', '\x00', '\x13'], # Has optional args now
        'disable_motors'     : ['\x02', '\x01', '\x00', '\x14'], # Has optional args now
        'circle_dance'       : ['\x02', '\x04', '\x00', '\x1C'], #
        'enable_safeties'    : ['\x02', '\x01', '\x00', '\x1E'], #
        'fall_protection'    : ['\x02', '\x02', '\x00', '\x15'], #
        'motor_protection'   : ['\x02', '\x02', '\x00', '\x16'], #
        'battery_protection' : ['\x02', '\x02', '\x00', '\x17'], #
        'buzz_prevention'    : ['\x02', '\x02', '\x00', '\x18'], #
        'lifelike_behaviour' : ['\x02', '\x02', '\x00', '\x1D'], #
        'clear_calibration'  : ['\x02', '\x01', '\x00', '\xFE'], #
        'save_calibration'   : ['\x02', '\x01', '\x00', '\xFF'], #
        'ros_command'        : ['\x03'],                         # Variable Length
        'set_param'          : ['\x02', '\x1F'], #
        'firmware_version'   : ['\x02', '\x01', '\x00', '\x20'], #
        'mute_serial'        : ['\x02', '\x01', '\x00', '\x21'], # OK
        'i2c_write'          : ['\x02'],                         #
        'gpio_write'         : ['\x02', '\x06', '\x00', '\x1A'], #
        'gpio_mode'          : ['\x02', '\x03', '\x00', '\x19'], #
    }

    def _pack(self, characters: List[str]) -> bytes:
        '''
        Pack characters list into a byte string
        Expects pre-packed chars, use chr or struct.pack to do this
        '''
        if self.debug:
            print(list(map(lambda x: '{}:{}'.format(x, type(x)), characters)))
        try:
            return ("".join(characters)).encode("latin-1")
        except UnicodeEncodeError:
            raise ArgumentOutOfRangeException('Argument(s) overflowed int')

    def _fixed_command(self, *args: int, **kwargs: int) -> bool:
        '''
        Send a command with a fixed (preknown) number of arguments
        that does not expect a response.
        NB!: Assumes data (args[2:]) is broken into 8-bit bytes
        '''
        cmd = args[1]
        opcode = self.CMD_OPCODES[cmd]
        datalen = ord(opcode[1]) + (ord(opcode[2]) << 8) - 1
        data = list(args[2:])
        if len(data) != datalen:
            raise TypeError('{} takes {} arguments but {} were given'
                            ''.format(cmd, datalen, len(data)))
        self.sock.send(self._pack(opcode + data))
        return True

    def _command(self, *args: int, **kwargs: int) -> bool:
        '''
        Pipes args down the socket whilst calculating the payload length
        Args:
            *args of length at least 3
        '''
        cmd = args[1]
        opcode = self.CMD_OPCODES[cmd][0]
        data = list(args[2:])
        datalen_lsb, datalen_msb = struct.pack('<H', len(data))
        payload = [opcode,
                   chr(datalen_lsb[0]),
                   chr(datalen_msb[0])] + data
        self.sock.send(self._pack(payload))
        return True

    def _toggle_command(self, *args: int, **kwargs: int) -> bool:
        '''
        Takes a python Boolean and toggles a switch on the board
        '''
        cmd = args[1]
        toggle = '\x01' if args[2] == True else '\x00'
        opcode = self.CMD_OPCODES[cmd]
        self.sock.send(self._pack(opcode + [toggle]))
        return args[2]

    def _simple_sensor(self, *args: int, **kwargs: int) -> int:
        '''
        Read a simple sensor and give its value
        Args:
            cmd
        '''
        cmd = args[1]
        self.sock.send(self._pack(self.CMD_OPCODES[cmd]))
        data = self.sock.recv(4)
        return struct.unpack('f', data)[0]

    def _select_sensor(self, *args: int, **kwargs: int) -> float:
        '''
        Read a sensor that takes an argument and give its value
        Args:
            cmd, index
        '''
        cmd = args[1]
        index = args[2]
        self.sock.send(self._pack(self.CMD_OPCODES[cmd] + [index]))
        data = self.sock.recv(4)
        return struct.unpack('f', data)[0]

    def _current_sensor(self, *args: int, **kwargs: int) -> float:
        '''
        Read current sensor that takes an argument and give its value
        Args:
            cmd, index
        '''
        cmd = args[1]
        index = args[2]
        pakd = self._pack(self.CMD_OPCODES[cmd] + [chr(int(index))])
        self.sock.send(pakd)
        data = self.sock.recv(4)
        return struct.unpack('f', data)[0]

    def _chatter(self, *args: int, **kwargs: int) -> bytes:
        '''
        Return chatter topic data (variable length)
        '''
        cmd = args[1]
        self.sock.send(self._pack(self.CMD_OPCODES[cmd]))
        data_length = self.sock.recv(1)[0]
        data = self.sock.recv(data_length)
        return data

    def _pack_uint16(self, num: int) -> Tuple[str,str]:
        '''
        Pack an unsigned 16 bit int into two 8 bit bytes, little-endian
        Returns:
            tuple(least-sig-byte, most-sig-byte)

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            h      short/uint16           integer        2
        '''
        try:
            data = struct.pack('<H', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return chr(data[0]), chr(data[1])

    def _pack_int16(self, num: int) -> Tuple[str,str]:
        '''
        Pack a signed 16 bit int into two 8 bit bytes, little-endian
        Returns:
            tuple(least-sig-byte, most-sig-byte)

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            h      short/uint16           integer        2
        '''
        try:
            data = struct.pack('<h', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return chr(data[0]), chr(data[1])

    def _pack_int32(self, num: int) -> Tuple[str,str,str,str]:
        '''
        Pack a signed 32 bit int into four 8 bit bytes, little-endian
        Returns:
            tuple(least-sig-byte, less-sig-byte, more-sig-byte, most-sig-byte)

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            i      int                    integer        4
        '''
        try:
            data = struct.pack('<i', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return (chr(data[0]),
                chr(data[1]),
                chr(data[2]),
                chr(data[3]))

    def _pack_uint8(self, num: int) -> str:
        '''
        Pack an unsigned 8 bit int into one 8 bit byte, little-endian
        Returns:
            bytes

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            B      unsigned char/uint8    integer        1
        '''
        try:
            data = struct.pack('<B', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return chr(data[0])

    def _pack_int8(self, num: int) -> str:
        '''
        Pack a signed 8 bit int into one 8 bit unsigned byte, little-endian
        Returns:
            bytes

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            b      signed char/int8       integer        1
        '''
        try:
            data = struct.pack('<b', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return chr(data[0])

    def _pack_float(self, num: float) -> Tuple[str,str,str,str]:
        '''
        Pack a float into four bit unsigned byte, little-endian
        Returns:
            tuple(least-sig-byte, less-sig-byte, more-sig-byte, most-sig-byte)

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            f      float                  float          4
        '''
        try:
            data = struct.pack('<f', float(num))
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return (chr(data[0]),
                chr(data[1]),
                chr(data[2]),
                chr(data[3]))

    def _execute(self, *args, **kwargs):
        '''
        Execute the command prescribed by the commands lookup table
        '''
        try:
            return self.COMMANDS_LUT[args[0]](self, *args, **kwargs)
        except KeyError:
            raise UnavailableCommandException("The command '{}' is not available "
                                              "from the {} client type"
                                              "".format(args[0], str(self.__class__.__name__)))
        except Exception as e:
            raise MartyCommandException(e)

    def _register_commands(self, handlers):
        '''
        Take {str:func} command names & handlers in a dict
        and register them with the COMMANDS_LUT
        '''
        self.COMMANDS_LUT = ClientGeneric.dict_merge(self.COMMANDS_LUT, handlers)

    def register_logging_callback(self, loggingCallback: Callable[[str],None]) -> None:
        return

    def get_interface_stats(self) -> Dict:
        return {}

    def preException(self, isFatal: bool) -> None:
        pass

    def get_test_output(self) -> dict:
        return ""



'''
Test communications with a Robotical RIC
'''
from threading import Thread
from typing import Callable, Dict, Union
import serial
import time
import copy
import logging
from .RICCommsBase import RICCommsBase
from .Exceptions import MartyConnectException

logger = logging.getLogger(__name__)

class RICCommsTest(RICCommsBase):
    def __init__(self) -> None:
        super().__init__()
        self._isOpen = True
        self.clearTestOutput()

    def __del__(self) -> None:
        try:
            self.close()
        except:
            pass

    def isOpen(self) -> bool:
        return self._isOpen

    def open(self, openParams: Dict) -> bool:
        self.commsParams.conn = openParams
        self.commsParams.fileTransfer = {"fileBlockMax": 5000, "fileXferSync": False}
        return True

    def close(self) -> None:
        self._isOpen = False

    def send(self, data: bytes) -> None:
        self.outputInfo["buf"].append(data)

    def getTestOutput(self) -> dict:
        dictCopy = copy.deepcopy(self.outputInfo)
        self.clearTestOutput()
        return dictCopy

    def clearTestOutput(self) -> None:
        self.outputInfo: Dict[str, list] = {"buf":[]}




import socket
import logging
from typing import Callable, Union
from .WebSocketFrame import WebSocketFrame

logger = logging.getLogger(__name__)

class WebSocket():

    maxSocketBytes = 2000
    MAX_RX_PREUPGRADE_LEN = 2000
    SOCKET_AWAITING_UPGRADE = 0
    SOCKET_UPGRADED = 1

    def __init__(self, 
            onBinaryFrame: Callable[[bytes], None],
            onTextFrame: Callable[[str], None],
            onError: Callable[[str], None],
            ipAddrOrHostname: str,
            ipPort: int = 80,
            wsPath: str = "/ws",
            timeout: float = 30.0) -> None:
        self.ipAddr = socket.gethostbyname(ipAddrOrHostname)
        self.wsPath = wsPath
        self.ipPort = ipPort
        self.timeout = timeout
        self.sock = None
        self.wsFrameCodec = WebSocketFrame()
        self.onBinaryFrame = onBinaryFrame
        self.onTextFrame = onTextFrame
        self.onError = onError
        self.socketState = self.SOCKET_AWAITING_UPGRADE
        self.rxPreUpgrade = bytearray()

    def __del__(self) -> None:
        self.close()

    def open(self) -> bool:
        # Open socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(self.timeout)
        self.sock.connect((self.ipAddr, self.ipPort))
        # Initiate upgrade
        self._sendUpgradeReq()

    def writeBinary(self, inFrame: bytes) -> int:
        if not self.sock:
            return 0
        frame = WebSocketFrame.encode(inFrame, False, WebSocketFrame.OPCODE_BINARY, True)
        # logger.debug(f"WebSocket write {''.join('{:02x}'.format(x) for x in frame)}") 
        return self.sock.send(frame)

    def close(self) -> None:
        if self.sock:
            self.sock.close()
        self.sock = None

    def _sendUpgradeReq(self) -> None:
        if not self.sock:
            return
        headerStr = f"GET {self.wsPath} HTTP/1.1\r\n" + \
            "Connection: upgrade\r\n" + \
            "Upgrade: websocket\r\n" + \
            "\r\n"
        self.sock.send(headerStr.encode())

    def service(self) -> None:
        # Get any data
        rxData = self.sock.recv(self.maxSocketBytes)
        # Check state of socket connection
        if self.socketState == self.SOCKET_AWAITING_UPGRADE:
            self.rxPreUpgrade += rxData
            # Check for upgrade header complete
            headerEndPos = self.rxPreUpgrade.find(b"\r\n\r\n")
            if headerEndPos > 0:
                # Check upgrade ok
                if b"Sec-WebSocket-Accept" in self.rxPreUpgrade:
                    self.socketState = self.SOCKET_UPGRADED
                    self.wsFrameCodec.addDataToDecode(self.rxPreUpgrade[headerEndPos+4:])
                    # logger.debug("WebSocket upgraded")
            elif len(self.rxPreUpgrade) > self.MAX_RX_PREUPGRADE_LEN:
                self.rxPreUpgrade.clear()
        else:
            self.wsFrameCodec.addDataToDecode(rxData)
            checkForData = True
            while checkForData:
                checkForData = False
                # Check for actions required
                if self.wsFrameCodec.getPongRequired():
                    self._sendPong()
                    checkForData = True
                binaryFrame = self.wsFrameCodec.getBinaryMsg()
                if binaryFrame:
                    checkForData = True
                    # logger.debug(f"WebSocket proc binary len {len(binaryFrame)}")
                    if self.onBinaryFrame:
                        self.onBinaryFrame(binaryFrame)
                textFrame = self.wsFrameCodec.getTextMsg()
                if textFrame and self.onTextFrame:
                    checkForData = True
                    self.onTextFrame(textFrame)

    def _sendPong(self) -> None:
        if not self.sock:
            return 0
        frame = WebSocketFrame.encode(self.wsFrameCodec.getPongData(),
                    False, WebSocketFrame.OPCODE_PONG, True)
        # logger.debug(f"WebSocket pong {''.join('{:02x}'.format(x) for x in self.wsFrameCodec.getPongData())}") 
        return self.sock.send(frame)



import time

class RateAverager:
    
    def __init__(self, windowSizeMinSecs = 1):
        self.sampleCount = 0
        self.windowSizeMinSecs = windowSizeMinSecs
        self.lastCalcSecs = time.time()
        self.prevVal = 0
        self.totalCount = 0

    def addSample(self):
        self.sampleCount += 1
        self.totalCount += 1

    def getAvg(self):
        if self.lastCalcSecs + self.windowSizeMinSecs < time.time():
            rsltVal = self.sampleCount / (time.time() - self.lastCalcSecs)
            self.lastCalcSecs = time.time()
            self.sampleCount = 0
            self.prevVal = rsltVal
            return rsltVal
        return self.prevVal

    def getTotal(self):
        return self.totalCount




class UnavailableCommandException(Exception):
    pass

class MartyCommandException(Exception):
    pass

class MartyConfigException(Exception):
    pass

class MartyConnectException(Exception):
    pass

class ArgumentOutOfRangeException(Exception):
    pass

class UnavailableClientTypeException(Exception):
    pass

class MartyTransferException(Exception):
    pass




'''
RICProtocols
'''
import json
import logging
from typing import Dict, Tuple, Union

logger = logging.getLogger(__name__)

class DecodedMsg:
    '''
    DecodedMsg
    A message decoded from RIC
    '''
    def __init__(self) -> None:
        self.errMsg = ""
        self.msgNum = None
        self.protocolID = None
        self.isText = None
        self.payload = None
        self.restType = None
        self.msgTypeCode = None

    def setError(self, errMsg: str) -> None:
        self.errMsg = errMsg

    def setMsgNum(self, msgNum: int) -> None:
        self.msgNum = msgNum

    def setProtocol(self, protocolID: int) -> None:
        self.protocolID = protocolID

    def setMsgTypeCode(self, msgTypeCode: int) -> None:
        self.msgTypeCode = msgTypeCode

    def setRESTElemCode(self, restType: int) -> None:
        self.restType = restType

    def setPayload(self, isText: bool, payload: bytes) -> None:
        self.isText = isText
        self.payload = payload

    def getJSONDict(self) -> Dict:
        msgContent = {}
        if self.isText:
            try:
                termPos = self.payload.find("\0")
                frameJson = self.payload
                if termPos >= 0:
                    frameJson = self.payload[0:termPos]
                    binData = self.payload[termPos+1:len(self.payload)-1]
                else:
                    return msgContent
                msgContent = json.loads(frameJson)
            except Exception as excp:
                logging.debug("RICProtocols getJSONDict failed to extract JSON from", 
                            self.payload, "error", excp)
        # logging.debug("msgContent ", msgContent)
        return msgContent

    def toString(self) -> str:
        msgStr = ""
        if self.msgNum is not None:
            msgStr += f"MsgNum: {self.msgNum} "
        if self.protocolID is not None:
            msgStr += "Protocol:"
            if self.protocolID == RICProtocols.PROTOCOL_RICREST:
                msgStr += "RICREST"
            elif self.protocolID == RICProtocols.PROTOCOL_M1SC:
                msgStr += "M1SC"
            elif self.protocolID == RICProtocols.PROTOCOL_ROSSERIAL:
                msgStr += "ROSSERIAL"
            else:
                msgStr += f"OTHER {self.protocolID}"
            msgStr += " "
        if self.msgTypeCode is not None:
            msgStr += "MsgType:"
            if self.msgTypeCode >= 0 and self.msgTypeCode < len(RICProtocols.MSG_TYPE_STRS):
                msgStr += RICProtocols.MSG_TYPE_STRS[self.msgTypeCode]
            else:
                msgStr += f"OTHER {self.msgTypeCode}"
            msgStr += " "
        if self.restType is not None:
            msgStr += "Type:"
            if self.restType >= 0 and self.restType < len(RICProtocols.REST_COMMANDS):
                msgStr += RICProtocols.REST_COMMANDS[self.restType]
            else:
                msgStr += f"OTHER {self.restType}"
            msgStr += " "
        if self.isText is not None:
            if self.isText:
                msgStr += "Payload:" + self.payload
            else:
                msgStr += f"PayloadLen: {len(self.payload)} "
        return msgStr

class RICProtocols:
    '''
    RICProtocols
    Implements protocols used by RIC including RICSerial and RICREST
    '''
    # Robotical RIC Protocol definitions
    MSG_TYPE_COMMAND = 0x00
    PROTOCOL_ROSSERIAL = 0x00
    PROTOCOL_M1SC = 0x01
    PROTOCOL_RICREST = 0x02
    RICREST_ELEM_CODE_URL = 0x00
    RICREST_ELEM_CODE_JSON = 0x01
    RICREST_ELEM_CODE_BODY = 0x02
    RICREST_ELEM_CODE_CMD_FRAME = 0x03
    RICREST_ELEM_CODE_FILE_BLOCK = 0x04
    REST_COMMANDS = ["url", "json", "body", "cmd", "fileBlk"]
    MSG_TYPE_COMMAND = 0x00
    MSG_TYPE_RESPONSE = 0x01
    MSG_TYPE_PUBLISH = 0x02
    MSG_TYPE_REPORT = 0x03
    MSG_TYPE_STRS = ["cmd", "resp", "publish", "report"]

    def __init__(self) -> None:
        self.ricSerialMsgNum = 1
        pass

    def encodeRICRESTURL(self, cmdStr: str) -> Tuple[bytes, int]:
        # RICSerial URL
        msgNum = self.ricSerialMsgNum
        cmdFrame = bytearray([msgNum, self.MSG_TYPE_COMMAND + self.PROTOCOL_RICREST, self.RICREST_ELEM_CODE_URL])
        cmdFrame += cmdStr.encode() + b"\0"
        self.ricSerialMsgNum += 1
        if self.ricSerialMsgNum > 255:
            self.ricSerialMsgNum = 1
        return cmdFrame, msgNum

    def encodeRICRESTCmdFrame(self, cmdStr: Union[str,bytes], payload: Union[bytes, str] = None) -> Tuple[bytes, int]:
        # RICSerial command frame
        msgNum = self.ricSerialMsgNum
        cmdFrame = bytearray([msgNum, self.MSG_TYPE_COMMAND + self.PROTOCOL_RICREST, self.RICREST_ELEM_CODE_CMD_FRAME])
        if type(cmdStr) is str:
            cmdFrame += cmdStr.encode()
        else:
            cmdFrame += cmdStr
        if cmdFrame[-1] != b"\0":
            cmdFrame = cmdFrame + b"\0"
        if payload is not None:
            if type(payload) is str:
                payload = payload.encode()
            cmdFrame += payload
        self.ricSerialMsgNum += 1
        if self.ricSerialMsgNum > 255:
            self.ricSerialMsgNum = 1
        return cmdFrame, msgNum

    def encodeRICRESTFileBlock(self, cmdBuf: bytes) -> Tuple[bytes, int]:
        # RICSerial file block - not numbered
        cmdFrame = bytearray([0, self.MSG_TYPE_COMMAND + self.PROTOCOL_RICREST, self.RICREST_ELEM_CODE_FILE_BLOCK])
        cmdFrame += cmdBuf
        return cmdFrame, 0

    def decodeRICFrame(self, fr: bytes) -> DecodedMsg:
        msg = DecodedMsg()
        if len(fr) < 2:
            msg.setError(f"Frame too short {len(fr)} bytes")
        else:
            # Decode header
            msgNum = fr[0]
            msg.setMsgNum(msgNum)
            protocol = fr[1] & 0x3f
            msg.setProtocol(protocol)
            msgTypeCode = fr[1] >> 6
            msg.setMsgTypeCode(msgTypeCode)
            if protocol == self.PROTOCOL_RICREST:
                restElemCode = fr[2]
                msg.setRESTElemCode(restElemCode)
                if restElemCode == self.RICREST_ELEM_CODE_URL or restElemCode == self.RICREST_ELEM_CODE_JSON:
                    msg.setPayload(True, fr[3:].decode('ascii'))
                    msg.payload = msg.payload.rstrip('\x00')
                else:
                    msg.setPayload(False, fr[3:])
                # logging.debug(f"RICREST {RICProtocols.MSG_TYPE_STRS[msgTypeCode]} msgNum {msgNum} {fr.hex()}")
            elif protocol == self.PROTOCOL_ROSSERIAL:
                msg.setPayload(False, fr[2:])
            else:
                logging.debug(f"RICProtocols Unknown frame received {fr}")
        return msg



'''
Parameters used for communication
'''
class RICCommsParams:

    def __init__(self, connParams: dict = {}, fileTransferParams: dict = {}) -> None:
        self.connParams = connParams
        self.fileTransferParams = fileTransferParams

    @property
    def conn(self) -> dict:
        return self.connParams

    @conn.setter
    def conn(self, connParams: dict) -> None:
        self.connParams = connParams

    @property
    def fileTransfer(self) -> dict:
        return self.fileTransferParams

    @fileTransfer.setter
    def fileTransfer(self, fileTransferParams: dict) -> None:
        self.fileTransferParams = fileTransferParams



'''
Serial communications with a Robotical RIC
'''
import logging
import serial
from serial.serialutil import SerialException
import serial.tools.list_ports
from threading import Thread
import time
from typing import Dict, List
from warnings import warn

from .RICCommsBase import RICCommsBase
from .LikeHDLC import LikeHDLC
from .ProtocolOverAscii import ProtocolOverAscii
from .Exceptions import MartyConnectException

logger = logging.getLogger(__name__)


class RICCommsSerial(RICCommsBase):
    '''
    RICCommsSerial
    Provides a serial interface for RIC communications
    '''
    def __init__(self) -> None:
        '''
        Initialise RICCommsSerial
        '''
        super().__init__()
        self._isOpen = False
        self.serialReaderThread: Thread = None
        self.serialDevice: serial.Serial = None
        self.serialThreadEnabled = False
        self._hdlc = LikeHDLC(self._onHDLCFrame, self._onHDLCError)
        self.overAscii = False
        self.serialLogLine = ""
        self.serialPortErrors = 0
        self.baudRateAlternates = [2000000, 115200]
        self.curBaudRate = 0
        self.baudRateCheckedOk = False
        self.msgRespTimeoutSecs = 1.5
        self.numMsgsTimedOutSinceRateChange = 0
        self.numMsgsTimedOutBeforeRateChange = 1

    def __del__(self) -> None:
        '''
        Destructor
        '''
        try:
            self.close()
        except:
            pass

    @classmethod
    def detect_rics(cls) -> List[str]:
        serial_ports = serial.tools.list_ports.comports()
        possible_rics = [port.device for port in serial_ports
                         if "CP210" in port.description]
        return possible_rics

    def isOpen(self) -> bool:
        '''
        Check if comms open
        Returns:
            True if comms open
        '''
        return self._isOpen

    def open(self, openParams: Dict) -> bool:
        '''
        Open connection
        Protocol can be plain (if the port is not used for any other purpose) or
                 overascii (if the port is also used for logging information)
        Args:
            openParams: dict containing params used to open the connection, may include
                        "serialPort" (will be auto-detected if missing/empty),
                        "serialBaud",
                        "ifType" == "plain" or "overascii",
                        "asciiEscapes"
        Returns:
            True if open succeeded or already open
        Throws:
            MartyConnectException: if connection cannot be opened
        '''
        # Check not already open
        if self._isOpen:
            return True

        # Get params
        self.commsParams.conn = openParams
        self.commsParams.fileTransfer = {"fileBlockMax": 5000, "fileXferSync": False}
        serialPort = openParams.get("serialPort", "")
        serialBaud = openParams.get("serialBaud", 2000000)
        self.overAscii = openParams.get("ifType", "plain") != "plain"
        if self.overAscii:
            self.protocolOverAscii = ProtocolOverAscii()
        hdlcAsciiEscapes = openParams.get("asciiEscapes", False)

        # Detect serial port with correct chip code
        rics = self.detect_rics()
        if not serialPort:
            if len(rics) == 0:
                raise MartyConnectException(
                    "No serial (COM) port provided and no Martys detected."
                )
            if len(rics) > 1:
                warn(f"Multiple Martys detected ({rics}). Connecting to {rics[0]}.",
                     stacklevel=0)
            serialPort = rics[0]

        # Try to open serial port
        try:
            self.baudRateCheckedOk = False
            self.curBaudRate = serialBaud
            self.serialDevice = serial.Serial(port=None, baudrate=serialBaud)
            self.serialDevice.port = serialPort
            self.serialDevice.rts = 0
            self.serialDevice.dtr = 0
            self.serialDevice.dsrdtr = False
            self.serialDevice.open()
        except Exception as excp:
            error_message = f"Serial port problem on {self.serialDevice.port}. Try "\
                            f"connecting to one of the following ports instead: {rics}"
            raise MartyConnectException(error_message) from excp

        # Configure HDLC
        self._hdlc.setAsciiEscapes(hdlcAsciiEscapes)
        
        # Start receive loop
        self.serialThreadEnabled = True
        self.serialReaderThread = Thread(target=self._serialRxLoop)
        self.serialReaderThread.daemon = True
        self.serialReaderThread.start()
        self._isOpen = True
        return True

    def close(self) -> None:
        '''
        Close serial port
        '''
        if not self._isOpen:
            return
        # Stop thread function
        if self.serialReaderThread is not None:
            self.serialThreadEnabled = False
            time.sleep(0.01)
            self.serialReaderThread.join()
            self.serialReaderThread = None
        # Close port
        if self.serialDevice is not None:
            self.serialDevice.close()
            self.serialDevice = None
        self._isOpen = False

    def send(self, data: bytes) -> None:
        '''
        Send data to serial port
        Args:
            data: bytes to send on serial port
        Returns:
            none
        Throws:
            MartyConnectException: if the connection has an error
        '''
        # logger.debug(f"Sending to IF len {len(bytesToSend)} {str(bytesToSend)}")
        hdlcEncoded = self._hdlc.encode(data)
        try:
            if self.overAscii:
                encodedFrame = ProtocolOverAscii.encode(hdlcEncoded)
                self._sendBytesToIF(encodedFrame)
                # logger.debug(f"send {encodedFrame.hex()}")
            else:
                self._sendBytesToIF(hdlcEncoded)
        except Exception as excp:
            raise MartyConnectException("Serial send problem") from excp

    def _serialRxLoop(self) -> None:
        '''
        Thread function used to process serial port data
        '''
        while self.serialThreadEnabled:
            i = self.serialDevice.in_waiting
            if i < 1:
                time.sleep(0.001)
                continue
            byt = self.serialDevice.read(i)
            for b in byt:
                if self.overAscii:
                    if b >= 128:
                        decodedVal = self.protocolOverAscii.decodeByte(b)
                        if decodedVal >= 0:
                            # logger.debug(f"{decodedVal:02x}")
                            self._hdlc.decodeData(decodedVal)
                    else:
                        if b == 0x0a:
                            if self.logLineCB:
                                self.logLineCB(self.serialLogLine)
                            self.serialLogLine = ""
                        else:
                            self.serialLogLine += chr(b)
                else:
                    self._hdlc.decodeData(b)
            # logger.debug(f"CommsSerial rx {byt.hex()}")
        # logger.debug("Exiting serialRxLoop")

    def _onHDLCFrame(self, frame: bytes) -> None:
        if self.rxFrameCB is not None:
            self.rxFrameCB(frame)

    def _onHDLCError(self) -> None:
        pass

    def _sendBytesToIF(self, bytesToSend: bytes) -> None:
        # logger.debug(f"Sending to IF len {len(bytesToSend)} {str(bytesToSend)}")
        if not self._isOpen:
            return
        # logger.debug(f"CommsSerial sendBytesToIF {''.join('{:02x}'.format(x) for x in bytesToSend)}")
        if self.serialDevice is not None:
            try:
                self.serialDevice.write(bytesToSend)
            except SerialException as excp:
                # The port has connected but now failing to send characters
                # This may sort itself out or require user intervention
                self.serialPortErrors += 1
                pass

    def getTestOutput(self) -> dict:
        return {}

    def getMsgRespTimeoutSecs(self, defaultValue):
        '''
        Get the timeout for a message response
        '''
        return self.msgRespTimeoutSecs

    def hintMsgTimeout(self, numTimedOut):
        '''
        Hint that messages are timing out from upper layers
        '''
        # Check if we've already found the right baud rate
        if not self.baudRateCheckedOk:
            # Count failed messages since last rate change
            self.numMsgsTimedOutSinceRateChange += numTimedOut
            if self.numMsgsTimedOutSinceRateChange >= self.numMsgsTimedOutBeforeRateChange:
                # Enough messages timed out since last rate change, so try next rate
                self._skipToNextBaudRate()
                self.numMsgsTimedOutSinceRateChange = 0
                # Back off the next rate change by a factor of 2
                self.numMsgsTimedOutBeforeRateChange *= 2

    def _skipToNextBaudRate(self):
        '''
        Skip to the next baud rate in the list
        '''
        if self.curBaudRate in self.baudRateAlternates:
            curBaudIdx = self.baudRateAlternates.index(self.curBaudRate)
            self.curBaudRate = self.baudRateAlternates[(curBaudIdx + 1) % len(self.baudRateAlternates)]
        else:
            self.curBaudRate = self.baudRateAlternates[0]
        self.serialDevice.baudrate = self.curBaudRate
        logger.debug(f"_skipToNextBaudRate comms failing - changing baud-rate to {self.curBaudRate}")



import sys
import logging

import pathlib
cur_path = pathlib.Path(__file__).parent.resolve()
sys.path.insert(0, str(cur_path.parent.parent.resolve()))

from martypy import Marty
import time

# Setup logging
logging.basicConfig(format='%(levelname)s: %(asctime)s %(funcName)s(%(lineno)d) -- %(message)s', level=logging.DEBUG)
logger = logging.getLogger("testBaudRateSelection")

mm = Marty("usb")
# try:
# mm = Marty("wifi://192.168.86.81")
# except Exception as excp:
#     print(f"Couldn't connect to marty")
#     exit()

# mm.eyes(100)
# mm.eyes(0)

logger.info(f"RIC version info 1: {mm.get_system_info()}")
mm.circle_dance()
logger.info(f"RIC version info 2: {mm.get_system_info()}")
time.sleep(5)
logger.info(f"RIC version info 3: {mm.get_system_info()}")

mm.close()



'''
WiFi communications with a Robotical RIC
'''
from threading import Thread
from typing import Callable, Dict, Union
import time
import logging
from .RICCommsBase import RICCommsBase
from .LikeHDLC import LikeHDLC
from .Exceptions import MartyConnectException
from .WebSocket import WebSocket

logger = logging.getLogger(__name__)

class RICCommsWiFi(RICCommsBase):
    '''
    RICCommsWiFi
    Provides an interface for RIC communications
    '''
    def __init__(self) -> None:
        '''
        Initialise RICCommsWiFi
        '''
        super().__init__()
        self._isOpen = False
        self.webSocketThread: Thread = None
        self.webSocket: WebSocket = None
        self.webSocketThreadEnabled = False
        self._hdlc = LikeHDLC(self._onHDLCFrame, self._onHDLCError)
        self.socketErrors = 0

    def __del__(self) -> None:
        '''
        Destructor
        '''
        try:
            self.close()
        except:
            pass

    def isOpen(self) -> bool:
        '''
        Check if comms open
        Returns:
            True if comms open
        '''
        return self._isOpen

    def open(self, openParams: Dict) -> bool:
        '''
        Open connection
        Args:
            openParams: dict containing params used to open the connection, may include
                        "ipAddrOrHostname", 
                        "ipPort",
                        "wsPath",
                        "asciiEscapes"
        Returns:
            True if open succeeded or already open
        Throws:
            MartyConnectException: if a connection cannot be opened
        '''
        # Check not already open
        if self._isOpen:
            return True

        # Get params
        self.commsParams.conn = openParams
        self.commsParams.fileTransfer = {"fileBlockMax": 5000, "fileXferSync": False}
        ipAddrOrHostname = openParams.get("ipAddrOrHostname", "")
        ipPort = openParams.get("ipPort", 80)
        wsPath = openParams.get("wsPath", "/ws")
        hdlcAsciiEscapes = openParams.get("asciiEscapes", False)

        # Validate
        if len(ipAddrOrHostname) == 0:
            return False

        # Open websocket
        try:
            self.webSocket = WebSocket(self._onWSBinaryFrame, 
                        self._onWSTextFrame, 
                        self._onWSError, 
                        ipAddrOrHostname, ipPort, wsPath)
            self.webSocket.open()
        except Exception as excp:
            raise MartyConnectException("Websocket problem") from excp

        # Configure HDLC
        self._hdlc.setAsciiEscapes(hdlcAsciiEscapes)

        # Start receive loop
        self.webSocketThreadEnabled = True
        self.webSocketThread = Thread(target=self._webSocketThreadFn)
        self.webSocketThread.daemon = True
        self.webSocketThread.start()
        self._isOpen = True
        return True

    def close(self) -> None:
        '''
        Close connection
        '''
        if not self._isOpen:
            return
        # Stop thread function
        if self.webSocketThread is not None:
            self.webSocketThreadEnabled = False
            time.sleep(0.01)
            self.webSocketThread.join()
            self.webSocketThread = None
        # Close socket
        if self.webSocket is not None:
            self.webSocket.close()
            self.webSocket = None
        self._isOpen = False

    def send(self, data: bytes) -> None:
        '''
        Send data
        Args:
            data: bytes to send
        Returns:
            none
        Throws:
            MartyConnectException: if the connection has an error
        '''
        # logger.debug(f"WiFi send len {len(data)} {''.join('{:02x}'.format(x) for x in data)}")
        hdlcEncoded = self._hdlc.encode(data)
        try:
            self._sendBytesToIF(hdlcEncoded)
        except Exception as excp:
            raise MartyConnectException("Connection send problem") from excp

    def _onHDLCFrame(self, frame: bytes) -> None:
        if self.rxFrameCB is not None:
            self.rxFrameCB(frame)
        
    def _onHDLCError(self) -> None:
        pass
        
    def _sendBytesToIF(self, bytesToSend: bytes) -> None:
        # logger.debug(f"Sending to IF len {len(bytesToSend)} {str(bytesToSend)}")
        if not self._isOpen:
            return
        # logger.debug(f"CommsWiFi sendBytesToIF {''.join('{:02x}'.format(x) for x in bytesToSend)}")
        if self.webSocket is not None:
            try:
                self.webSocket.writeBinary(bytesToSend)
            except Exception as excp:
                # The socket has connected but now failing to send characters
                # This may sort itself out or require user intervention
                self.socketErrors += 1
                pass

    def _webSocketThreadFn(self) -> None:
        '''
        Thread function used to process websocket data
        '''
        while self.webSocketThreadEnabled:
            try:
                self.webSocket.service()
            except OSError as excp:
                logger.debug(f"webSocket problem {excp}")
                self.webSocketThreadEnabled = False
            except Exception as excp:
                logger.debug(f"WebSocket exception {excp}")
            time.sleep(0.001)
        logger.debug("Exiting WebSocket thread")
        self._isOpen = False

    def _onWSBinaryFrame(self, rxFrame: bytes) -> None:
        # logger.debug(f"webSocketRx {''.join('{:02x}'.format(x) for x in rxFrame)}")
        for rxByte in rxFrame:
            self._hdlc.decodeData(rxByte)

    def _onWSTextFrame(self, rxFrame: str) -> None:
        logger.debug(f"webSocketRx TEXT UNEXPECTED {rxFrame}")

    def _onWSError(self, err: str) -> None:
        logger.debug(f"CommsWiFi WS error {err}")

    def getTestOutput(self) -> dict:
        return {}
        